
module Exp
    ( 
    modInv,
    integer2double,
    nfullrankha,
    nfullrankhb,
    invh,
    backsub,
    get_zfa,
    invhp,
    get_zf,
    gen_prezf,
    fget_cx_nc_last,
    decode_c
    ) where


import Data.List.Split
import Param
import System.IO  
import Mtrx
import Control.Monad
import Data.Matrix
import OpenSSL
import OpenSSL.BN
import File
import Data.Bits
import Decode
import Numeric.LinearAlgebra.HMatrix

decode_c :: Parameters -> IO Integer
decode_c param = do c1 <- fget_cx_ncoo param
                    c2 <- fget_fc param
                    zf <- gen_prezf param
                    c <- fget_cx_nc_last param
                    if (mod (((Data.Matrix.toList (multStd (c1 <|> c2) zf))!!0)+c ) (getq param)  )*2 < (getq param) then return $ 0 else return $ 1


fget_cx_nc_last :: Parameters -> IO Integer
fget_cx_nc_last param = do fcx <- fget_cx param
                           return $ (last (Data.Matrix.toList fcx))
--code from https://rosettacode.org/wiki/Modular_inverse#Haskell
-- Extended Euclidean algorithm.  Given non-negative a and b, return x, y and g
-- such that ax + by = g, where g = gcd(a,b).  Note that x or y may be negative.
gen_prezf :: Parameters -> IO (Data.Matrix.Matrix (Integer))
gen_prezf param = do zf1 <-  get_zfa param 
                     zf2 <- fget_zf2 param
                     return $ (zf1 <-> zf2)


--fget_zf2 :: Parameters -> IO (Matrix (Integer))
get_zfa :: Parameters -> IO (Data.Matrix.Matrix (Integer))
get_zfa param = do new_u <- gen_prezf1 param
                   hinv <- invh param
                   r <-fget_r param
                   return $ (safemodmat param (multStd r (ginvz param (safemodmat param (multStd hinv new_u )))))
get_zf :: Parameters -> IO (Data.Matrix.Matrix (Integer))
get_zf param = do zf2 <- fget_zf2 param
                  zfa <- get_zfa param
                  return $ (zfa <-> zf2)

gcdExt a 0 = (1, 0, a)
gcdExt a b = let (q, r) = a `quotRem` b
                 (s, t, g) = gcdExt b r
             in (t, s - q * t, g)
 
-- Given a and m, return Just x such that ax = 1 mod m.  If there is no such x
-- return Nothing.
modInv a m = let (i, _, g) = gcdExt a m
             in if g == 1 then (mkPos i) else 0
  where mkPos x = if x < 0 then x + m else x

integer2double :: (Data.Matrix.Matrix (Integer)) -> (Data.Matrix.Matrix (Double))
integer2double mati = (Data.Matrix.fromList (nrows mati) (ncols mati) (i2darr (Data.Matrix.toList mati)))

fgen_ha :: Parameters -> IO()
fgen_ha param = genrandintapm param "h" ((getn param)*(getn param))

getranka :: Parameters -> IO Integer
getranka param = do h <- fget_h param
                    let hrank = (rank (Numeric.LinearAlgebra.HMatrix.matrix (ncols h) (Data.Matrix.toList (integer2double h))))
                    return (toInteger hrank)
getrankb :: Parameters -> IO Bool
getrankb param = do hrank <- getranka param 
                    h <- (fget_h param)
                    if (hrank== (toInteger (nrows h))) then return True else return False

nfullrankha param = do fgen_ha param
                       hrank <- getranka param
                       h <- (fget_h param) 
                       if (hrank== (toInteger (nrows h))) then (nfullrankha param) else (nfullrankha param)
--out <- stringiointnpm param 10
--    print $ (strtointeger (lines out))

nfullrankhb param = do out <- stringiointnpm param ((getn param )*(getn param))
                       let arr = (strtointeger (lines out))
                       let hh = (Data.Matrix.fromList (fromInteger (getn param )) (fromInteger(getn param)) arr)
                       let hrank = (rank (Numeric.LinearAlgebra.HMatrix.matrix (ncols hh) (Data.Matrix.toList (integer2double hh))))
                       --if (hrank== ( (fromInteger (nrows hh))) then (fstorematrix hh "hh" ) else (nfullrankha param)
                       if ((toInteger hrank)< (toInteger ( (nrows hh)))) then  (nfullrankhb param) else (fstorematrix hh "hh" )
nfullrankh = False

fgen_hh :: Parameters -> IO()
fgen_hh param = loop (fgen_ha) 
    where loop idk | nfullrankh =  loop (fgen_ha)
                   | otherwise = return ()

i2darr :: [Integer] -> [Double]
i2darr arr = loop 0 []
    where loop iter output | iter < (length arr) = loop (iter+1) output++[(fromInteger (arr!!iter))]
                           | otherwise = output

invha :: Parameters -> Integer ->  IO (Data.Matrix.Matrix (Integer)) ->  IO (Data.Matrix.Matrix (Integer)) 
invha param piv hh = do h <- hh
                        return $ (invhmat param piv (h<|> (identity (nrows h))))



invhb :: Parameters -> Integer ->  IO (Data.Matrix.Matrix (Integer)) ->  IO (Data.Matrix.Matrix (Integer)) 
invhb param piv hh = do h <- hh
                        return $ (invhmatb param piv (h))

----This does forward eleminiation
invhp :: Parameters -> IO (Data.Matrix.Matrix (Integer)) 
invhp param = loop 1 (fget_h param)--(safemod param (invha param 1))
    where loop piv mat | piv ==1 = loop (piv+1) (safemod param (invha param piv mat))
                       | piv < ((getn param)) = loop (piv+1) (safemod param (invhb param piv mat))
                       | otherwise = mat
--(safemod param (invha param 1 (fget_h param))) --
--safemod param (invha param piv)


minus a b = a-b
--(rowVector (Data.Matrix.elementwise minus (Data.Matrix.getRow iter h ))     (scaleMatrix   (modInv  (getElem iter 1) (getq param)) (rowVector (Data.Matrix.getRow (1::Int) h)))      ))

idpiv :: Parameters -> Integer  -> (Data.Matrix.Matrix (Integer)) -> (Data.Matrix.Matrix (Integer)) 
idpiv param piv h =(switchRows (fromInteger (piv)) (fromInteger ((getn param)+1))(h <-> (scaleMatrix ((modInv (mod ((getElem (fromInteger piv) (fromInteger piv) h)) ((getq param)))(getq param))) (rowVector (Data.Matrix.getRow  (fromInteger (piv)) h))  )           ))

--(minorMatrix (fromInteger ((getn param)+1)) 0 (switchRows (fromInteger piv) (fromInteger ((getn param)+1)) (h <-> (scale (modInv (mod (getElem (fromInteger piv) 1 h) (getq param) ) (getq param) (rowVector (Data.Matrix.getRow  (fromInteger piv) h))))))

invhmat :: Parameters -> Integer  -> (Data.Matrix.Matrix (Integer)) -> (Data.Matrix.Matrix (Integer)) 
invhmat param piv h= loop (fromInteger (piv+1)) (rowVector (Data.Matrix.getRow  (fromInteger piv) (idpiv param piv h)))
    where loop iter matv | iter < ((nrows h)+1) = loop (iter+1) (matv <-> (elementwise minus (rowVector (Data.Matrix.getRow  (iter::Int) h))) (scaleMatrix  ((getElem iter (fromInteger piv) h)*(modInv (mod (getElem (fromInteger piv) (fromInteger piv) h) (getq param) ) (getq param))) (rowVector (Data.Matrix.getRow  (fromInteger piv) h))))
                         | otherwise = (Data.Matrix.fromList (nrows h) (ncols h)(Data.Matrix.toList (idpiv param  (toInteger (nrows matv)) matv)))
invhmatb :: Parameters -> Integer  -> (Data.Matrix.Matrix (Integer)) -> (Data.Matrix.Matrix (Integer)) 
invhmatb param piv h= loop (fromInteger (piv+1)) ((submatrix 1 (fromInteger (piv-1)) 1 (ncols h) h) <-> (rowVector (Data.Matrix.getRow  (fromInteger piv) (idpiv param piv h))))
    where loop iter matv | iter < ((nrows h)+1) = loop (iter+1) (matv <-> rowVector (Data.Matrix.getRow  (iter::Int) h)) 

--loop (iter+1) (matv <-> (elementwise minus (rowVector (Data.Matrix.getRow  (iter::Int) h))) (scaleMatrix  ((getElem iter (fromInteger piv) h)*(modInv (mod (getElem (fromInteger piv) (fromInteger piv) h) (getq param) ) (getq param))) (rowVector (Data.Matrix.getRow  (fromInteger piv) h))))
                         | otherwise = h--(Data.Matrix.fromList (nrows h) (ncols h)(Data.Matrix.toList (idpiv param  (toInteger (nrows matv)) matv)))
--invhmat param piv h= loop (fromInteger (piv+1)) (rowVector (Data.Matrix.getRow  (fromInteger piv) (idpiv param piv h)))
--    where loop iter matv | iter < ((nrows h)+1) = loop (iter+1) (matv <-> (elementwise minus (rowVector (Data.Matrix.getRow  (iter::Int) h))) (scaleMatrix  ((getElem iter 1 h)*(modInv (mod (getElem (fromInteger piv) 1 h) (getq param) ) (getq param))) (rowVector (Data.Matrix.getRow  (fromInteger piv) h))))
--                         | otherwise = (Data.Matrix.fromList (nrows h) (ncols h)(Data.Matrix.toList (idpiv param  (toInteger (nrows matv)) matv)))

--hinv <- invh param

backsub :: Parameters -> IO (Data.Matrix.Matrix (Integer)) 
backsub param = loop (getn param) (invhp param)--(safemod param (invha param 1))
    where loop piv mat | piv > 1= loop (piv-1) (safemod param (backsuba param piv mat))
                       | otherwise = mat

------------
backsuba :: Parameters -> Integer ->  IO (Data.Matrix.Matrix (Integer)) ->  IO (Data.Matrix.Matrix (Integer)) 
backsuba param piv iomat = do mat <- iomat
                              return $ (backsubb param piv mat)
-----------------

--((getElem (fromInteger iter) (fromInteger piv) mat)*(modInv (getElem (fromInteger piv) (fromInteger piv) mat) (getq param)  (getq param)) )

--(modInv (getElem (fromInteger piv) (fromInteger piv) mat) (getq param))

getmrow :: Integer ->(Data.Matrix.Matrix (Integer)) -> (Data.Matrix.Matrix (Integer)) 
getmrow iter mat = (rowVector (Data.Matrix.getRow  (fromInteger iter) (mat)))


--to do, make the backsub actually work haha
backsubb:: Parameters   -> Integer -> (Data.Matrix.Matrix (Integer)) -> (Data.Matrix.Matrix (Integer)) 
backsubb param piv mat= loop ((getn param)-1) (rowVector (Data.Matrix.getRow  (fromInteger (getn param)) (mat)))
    where loop iter hfe | iter > 0 = loop (fromInteger (iter-1)) (  (elementwise minus (getmrow iter mat) (scaleMatrix ((getElem (fromInteger iter) (fromInteger piv) mat)*(modInv (getElem (fromInteger piv) (fromInteger piv) mat) (getq param)))    (getmrow piv mat)))<->hfe)
                        | otherwise = hfe



invh :: Parameters -> IO (Data.Matrix.Matrix (Integer)) 
invh param = do out<- (backsub param)
                return (submatrix  1 (fromInteger ((getn param))) (fromInteger ((getn param)+1))  (ncols out) out)



