module Param
    ( Parameters (..)
    , getn
    , getm
    , getmb
    , getq
    , getw
    , getl
    , gen_parameters
    , genrandint
    , test
    , primesBetween
    , primesBetweena
    , log2
    ) where

import OpenSSL
import OpenSSL.BN
import Data.Numbers.Primes

--data Parameters = Integer Integer Integer Integer
--data Parameters = Parameters {n::Integer, m::Integer, m_bar::Integer, q::Integer}
data Parameters = Parameters Integer Integer Integer Integer deriving (Show)

getn :: Parameters -> Integer
getn (Parameters n _ _ _) = n
getm :: Parameters -> Integer
getm (Parameters _ m _ _) = m
getmb :: Parameters -> Integer
getmb (Parameters _ _ mb _) = mb
getq :: Parameters -> Integer
getq (Parameters _ _ _ q) = q
getw :: Parameters -> Integer
getw (Parameters n _ _ q) = n*(log2 q)
getl :: Parameters -> Integer
getl (Parameters n _ _ q) = (log2 q)
gen_parameters :: Integer -> Integer -> Parameters
gen_parameters n q= Parameters n (2*n*(log2 q)) (n*(log2 q)) q
--note this is setting m_bar=nl for simplicity, can be changed if desired

genrandint :: Integer -> IO Integer
genrandint n=randIntegerZeroToNMinusOne n

log2 :: Integer -> Integer
log2 n
    | n < 1     = error "agument of logarithm must be positive"
    | otherwise = go 0 1
      where
        go exponent prod
            | prod < n  = go (exponent + 1) (2*prod)
            | otherwise = exponent


test :: Integer -> Integer
test n = toInteger n


testa :: Integer -> Integer
testa n = toInteger n

--https://github.com/Answeror/spoj-hs/blob/master/PRIME1.hs
primesBetween :: Integer -> Integer ->[Integer]
primesBetween m n
    | m > n = []
    | otherwise = eratos ps xs
    where ps = (takeWhile (\x -> x*x <= n) primes)
          xs = dropWhile (<2) [m..n]
          eratos (p:ps) xs = eratos ps (xs `minus` [first, first+p..n])
            where first
                    | p < m = p*((m+p-1) `div` p)
                    | otherwise = p + p
          eratos _ xs = xs


minus (x:xs) (y:ys) = case (compare x y) of
                           LT -> x:minus xs (y:ys)
                           EQ -> minus xs ys
                           GT -> minus (x:xs) ys
minus xs _ = xs


primesBetweena :: Integer -> Integer ->Integer
primesBetweena m n
    | m > n = 0
    | otherwise = randelem (eratos ps xs)
    where ps = (takeWhile (\x -> x*x <= n) primes)
          xs = dropWhile (<2) [m..n]
          eratos (p:ps) xs = eratos ps (xs `minus` [first, first+p..n])
            where first
                    | p < m = p*((m+p-1) `div` p)
                    | otherwise = p + p
          eratos _ xs = xs

randelem :: [Integer]->Integer
randelem n 
    | (length n)<2 = head n
    | otherwise = head (drop ((length n)-1) n)




